# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from feaas import runners, storage


class InstanceTerminator(runners.Base):
    lock_name = "instance_terminator"

    def __init__(self, *args, **kwargs):
        super(InstanceTerminator, self).__init__(*args, **kwargs)
        self.init_locker(self.lock_name)

    def run(self):
        try:
            instance = self.get_instance()
            self.terminate_instance(instance)
        except storage.InstanceNotFoundError:
            pass

    def get_instance(self):
        self.locker.lock(self.lock_name)
        try:
            instance = self.storage.retrieve_instance(state="removed")
            instance.state = "terminating"
            self.storage.store_instance(instance)
            return instance
        finally:
            self.locker.unlock(self.lock_name)

    def terminate_instance(self, instance):
        self.locker.lock(self.lock_name)
        try:
            self.manager.terminate_instance(instance.name)
        finally:
            try:
                self.storage.remove_instance(instance.name)
            finally:
                self.locker.unlock(self.lock_name)



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import os
import sys
import time
import uuid

from feaas import managers, storage

from .cloudstack_client import CloudStack


class CloudStackManager(managers.BaseManager):

    def __init__(self, *args, **kwargs):
        super(CloudStackManager, self).__init__(*args, **kwargs)
        url = self.get_env("CLOUDSTACK_API_URL")
        key = self.get_env("CLOUDSTACK_API_KEY")
        secret_key = self.get_env("CLOUDSTACK_SECRET_KEY")
        self.client = CloudStack(url, key, secret_key)

    def get_env(self, name):
        try:
            return os.environ[name]
        except KeyError:
            raise MissConfigurationError("env var {} is required".format(name))

    def start_instance(self, name):
        instance = self.storage.retrieve_instance(name=name)
        self._add_units(instance, 1)
        return instance

    def terminate_instance(self, name):
        instance = self.storage.retrieve_instance(name=name)
        for unit in instance.units:
            self._destroy_vm(unit)
        return instance

    def physical_scale(self, instance, quantity):
        new_units = quantity - len(instance.units)
        if new_units < 0:
            return self._remove_units(instance, -1 * new_units)
        return self._add_units(instance, new_units)

    def _add_units(self, instance, quantity):
        units = []
        for i in xrange(quantity):
            unit = self._deploy_vm(instance)
            instance.add_unit(unit)
            units.append(unit)
        self.storage.store_instance(instance)
        return units

    def _deploy_vm(self, instance):
        secret = unicode(uuid.uuid4())
        group = os.environ.get("CLOUDSTACK_GROUP", "feaas")
        data = {
            "group": group,
            "displayname": "{}_{}".format(group, instance.name),
            "templateid": self.get_env("CLOUDSTACK_TEMPLATE_ID"),
            "zoneid": self.get_env("CLOUDSTACK_ZONE_ID"),
            "serviceofferingid": self.get_env("CLOUDSTACK_SERVICE_OFFERING_ID"),
            "userdata": self.client.encode_user_data(self.get_user_data(secret)),
        }
        project_id = os.environ.get("CLOUDSTACK_PROJECT_ID")
        if project_id:
            data["projectid"] = project_id
        network_id = os.environ.get("CLOUDSTACK_NETWORK_ID")
        if network_id:
            data["networkid"] = network_id
        vm_job = self.client.deployVirtualMachine(data)
        max_tries = int(os.environ.get("CLOUDSTACK_MAX_TRIES", 100))
        vm = self._wait_for_unit(vm_job, max_tries, project_id)
        return storage.Unit(id=vm["id"], dns_name=vm["nic"][0]["ipaddress"],
                            state="creating", secret=secret)

    def _wait_for_unit(self, vm_job, max_tries, project_id):
        status = 0
        tries = 0
        job_id = vm_job["jobid"]
        while tries < max_tries:
            result = self.client.queryAsyncJobResult({"jobid": job_id})
            status = result["jobstatus"]
            if status != 0:
                break
            time.sleep(1)
            tries += 1
        if status == 0:
            raise MaxTryExceededError(max_tries)
        data = {"id": vm_job["id"]}
        if project_id:
            data["projectid"] = project_id
        vms = self.client.listVirtualMachines(data)
        return vms["virtualmachine"][0]

    def _remove_units(self, instance, quantity):
        units = []
        for i in xrange(quantity):
            self._destroy_vm(instance.units[i])
            units.append(instance.units[i])
        for unit in units:
            instance.remove_unit(unit)
        self.storage.store_instance(instance)
        return units

    def _destroy_vm(self, unit):
        try:
            self.client.destroyVirtualMachine({"id": unit.id})
        except Exception as e:
            sys.stderr.write("[ERROR] Failed to terminate CloudStack VM: %s" %
                             " ".join([str(arg) for arg in e.args]))


class MissConfigurationError(Exception):
    pass


class MaxTryExceededError(Exception):

    def __init__(self, max_tries):
        self.max_tries = max_tries
        msg = "exceeded {0} tries".format(max_tries)
        super(MaxTryExceededError, self).__init__(msg)



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import functools
import os

import flask


def check_auth(auth):
    username = os.environ.get("API_USERNAME")
    password = os.environ.get("API_PASSWORD")
    if not username or not password:
        return True
    return auth and auth.username == username and auth.password == password


def required(fn):
    @functools.wraps(fn)
    def decorated(*args, **kwargs):
        auth = flask.request.authorization
        if not check_auth(auth):
            return "you do not have access to this resource", 401
        return fn(*args, **kwargs)
    return decorated



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from feaas import runners, storage


class InstanceScalator(runners.Base):
    lock_name = "instance_scalator"

    def __init__(self, *args, **kwargs):
        super(InstanceScalator, self).__init__(*args, **kwargs)
        self.init_locker(self.lock_name)

    def run(self):
        try:
            instance, job = self.get_job()
            if not job:
                return
            self.scale_instance(instance, job["quantity"])
            self.storage.finish_scale_job(job)
        except storage.InstanceNotFoundError:
            pass

    def get_job(self):
        self.locker.lock(self.lock_name)
        try:
            job = self.storage.get_scale_job()
            if not job:
                return None, None
            instance = self.storage.retrieve_instance(name=job["instance"],
                                                      check_liveness=True)
            if instance.state != "started":
                self.storage.reset_scale_job(job)
                return None, None
            instance.state = "scaling"
            self.storage.store_instance(instance)
            return instance, job
        except storage.InstanceNotFoundError:
            self.storage.finish_scale_job(job)
            raise
        finally:
            self.locker.unlock(self.lock_name)

    def scale_instance(self, instance, quantity):
        lock_name = "%s/%s" % (self.lock_name, instance.name)
        self.locker.init(lock_name)
        self.locker.lock(lock_name)
        try:
            try:
                self.manager.physical_scale(instance, quantity)
            finally:
                instance.state = "started"
                self.storage.store_instance(instance, save_units=False)
        finally:
            self.locker.unlock(lock_name)



#!/usr/bin/env python

# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import argparse
import os
import urllib2
import sys


class CommandNotFoundError(Exception):

    def __init__(self, name):
        super(Exception, self).__init__(name)
        self.name = name

    def __str__(self):
        return """command "{}" not found""".format(self.name)

    def __unicode__(self):
        return unicode(str(self))


def scale(args):
    instance, quantity = get_scale_args(args)
    result = proxy_request(instance, "/resources/{}/scale".format(instance),
                           body="quantity={}".format(quantity))
    if result.getcode() == 201:
        msg = "Instance successfully scaled to {} unit".format(quantity)
        if quantity > 1:
            msg += "s"
        sys.stdout.write(msg + "\n")
    else:
        msg = result.read().rstrip("\n")
        sys.stderr.write("ERROR: " + msg + "\n")
        sys.exit(1)


def get_scale_args(args):
    parser = argparse.ArgumentParser("scale")
    parser.add_argument("-i", "--instance")
    parser.add_argument("-n", "--quantity", type=int)
    parsed_args = parser.parse_args(args)
    if parsed_args.instance is None or parsed_args.quantity is None:
        parser.print_usage(sys.stderr)
        sys.exit(2)
    if parsed_args.quantity < 1:
        sys.stderr.write("quantity must be a positive integer\n")
        sys.exit(2)
    return parsed_args.instance, parsed_args.quantity


def get_env(name):
    env = os.environ.get(name)
    if not env:
        sys.stderr.write("ERROR: missing {}\n".format(name))
        sys.exit(2)
    return env


def proxy_request(instance_name, path, body=None):
    target = get_env("TSURU_TARGET").rstrip("/")
    token = get_env("TSURU_TOKEN")
    url = "{}/services/proxy/{}?callback={}".format(target, instance_name,
                                                    path)
    request = urllib2.Request(url)
    request.add_header("Authorization", "bearer " + token)
    if body:
        request.add_data(body)
    return urllib2.urlopen(request)


def get_command(name):
    commands = {
        "scale": scale,
    }
    command = commands.get(name)
    if not command:
        raise CommandNotFoundError(name)
    return command


def main(cmd, args):
    try:
        command = get_command(cmd)
        command(args)
    except CommandNotFoundError as e:
        sys.stderr.write(unicode(e) + u"\n")
        sys.exit(2)

if __name__ == "__main__":
    main(sys.argv[1], sys.argv[2:])



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import os
import urlparse
import uuid
import sys

from feaas import managers, storage


class EC2Manager(managers.BaseManager):

    def __init__(self, *args, **kwargs):
        super(EC2Manager, self).__init__(*args, **kwargs)
        self._connection = None

    @property
    def connection(self):
        if not self._connection:
            self._connection = self._connect()
        return self._connection

    def _connect(self):
        endpoint = os.environ.get("EC2_ENDPOINT", "https://ec2.sa-east-1.amazonaws.com")
        access_key = os.environ.get("EC2_ACCESS_KEY")
        secret_key = os.environ.get("EC2_SECRET_KEY")
        from boto import ec2
        url = urlparse.urlparse(endpoint)
        scheme = url.scheme
        host_port = url.netloc.split(":")
        host = host_port[0]
        if len(host_port) > 1:
            port = int(host_port[1])
        else:
            port = 80 if scheme == "http" else 443
        path = url.path or "/"
        region = ec2.RegionInfo(name="custom", endpoint=host)
        return ec2.EC2Connection(aws_access_key_id=access_key,
                                 aws_secret_access_key=secret_key,
                                 host=host, port=port,
                                 is_secure=scheme == "https",
                                 path=path, region=region)

    def start_instance(self, name):
        instance = self.storage.retrieve_instance(name=name)
        self._add_units(instance, 1)
        return instance

    def _run_unit(self):
        ami_id = os.environ.get("AMI_ID")
        subnet_id = os.environ.get("SUBNET_ID")
        secret = unicode(uuid.uuid4())
        reservation = self.connection.run_instances(image_id=ami_id,
                                                    subnet_id=subnet_id,
                                                    user_data=self._user_data(secret))
        ec2_instance = reservation.instances[0]
        return storage.Unit(id=ec2_instance.id, dns_name=ec2_instance.dns_name,
                            secret=secret, state="creating")

    def _user_data(self, secret):
        return self.get_user_data(secret)

    def terminate_instance(self, name):
        instance = self.storage.retrieve_instance(name=name)
        for unit in instance.units:
            self._terminate_unit(unit)
        return instance

    def _terminate_unit(self, unit):
        try:
            self.connection.terminate_instances(instance_ids=[unit.id])
        except Exception as e:
            sys.stderr.write("[ERROR] Failed to terminate EC2 instance: %s" %
                             " ".join([str(arg) for arg in e.args]))

    def physical_scale(self, instance, quantity):
        new_units = quantity - len(instance.units)
        if new_units < 0:
            return self._remove_units(instance, -1 * new_units)
        return self._add_units(instance, new_units)

    def _add_units(self, instance, quantity):
        units = []
        for i in xrange(quantity):
            unit = self._run_unit()
            instance.add_unit(unit)
            units.append(unit)
        self.storage.store_instance(instance)
        return units

    def _remove_units(self, instance, quantity):
        units = []
        for i in xrange(quantity):
            self._terminate_unit(instance.units[i])
            units.append(instance.units[i])
        for unit in units:
            instance.remove_unit(unit)
        self.storage.store_instance(instance)
        return units



# Copyright 2012 Kelcey Damage & Kraig Amador
# Source: http://goo.gl/KQUeMd

import base64
import hmac
import hashlib
import json
import urllib


class CloudStack(object):

    def __init__(self, api_url, api_key, secret):
        self.api_url = api_url
        self.api_key = api_key
        self.secret = secret

    def encode_user_data(self, data):
        return base64.b64encode(data)

    def request(self, args):
        args["apiKey"] = self.api_key
        self.params = []
        self._sort_request(args)
        self._create_signature()
        self._build_post_request()

    def _sort_request(self, args):
        keys = sorted(args.keys())
        for key in keys:
            self.params.append(key + "=" + urllib.quote_plus(args[key]))

    def _create_signature(self):
        self.query = "&".join(self.params)
        digest = hmac.new(self.secret, msg=self.query.lower(),
                          digestmod=hashlib.sha1).digest()
        self.signature = base64.b64encode(digest)

    def _build_post_request(self):
        self.query += "&signature=" + urllib.quote_plus(self.signature)
        self.value = self.api_url + "?" + self.query

    def __getattr__(self, name):
        def handler(*args, **kwargs):
            if kwargs:
                return self._make_request(name, kwargs)
            return self._make_request(name, args[0])
        return handler

    def _http_get(self, url):
        response = urllib.urlopen(url)
        return response.read()

    def _make_request(self, command, args):
        args["response"] = "json"
        args["command"] = command
        self.request(args)
        data = self._http_get(self.value)
        key = command.lower() + "response"
        return json.loads(data)[key]



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import inspect
import json
import os

from flask import Flask, Response, request

from . import auth, plugin, storage
from .managers import cloudstack, ec2

api = Flask(__name__)
api.debug = os.environ.get("API_DEBUG", "0") in ("True", "true", "1")

managers = {
    "ec2": ec2.EC2Manager,
    "cloudstack": cloudstack.CloudStackManager,
}


@api.route("/resources", methods=["POST"])
@auth.required
def add_instance():
    name = request.form.get("name")
    if not name:
        return "name is required", 400
    manager = get_manager()
    manager.new_instance(name)
    return "", 201


@api.route("/resources/<name>", methods=["DELETE"])
@auth.required
def remove_instance(name):
    manager = get_manager()
    try:
        manager.remove_instance(name)
    except storage.InstanceNotFoundError:
        return "Instance not found", 404
    return "", 200


@api.route("/resources/<name>", methods=["POST"])
@auth.required
def bind(name):
    app_host = request.form.get("app-host")
    if not app_host:
        return "app-host is required", 400
    manager = get_manager()
    try:
        manager.bind(name, app_host)
    except storage.InstanceNotFoundError:
        return "Instance not found", 404
    return Response(response="null", status=201,
                    mimetype="application/json")


@api.route("/resources/<name>/hostname/<host>", methods=["DELETE"])
@auth.required
def unbind(name, host):
    manager = get_manager()
    try:
        manager.unbind(name, host)
    except storage.InstanceNotFoundError:
        return "Instance not found", 404
    return "", 200


@api.route("/resources/<name>", methods=["GET"])
@auth.required
def info(name):
    manager = get_manager()
    try:
        info = manager.info(name)
        if "secret" in info:
            del info["secret"]
        return Response(response=json.dumps(info), status=200,
                        mimetype="application/json")
    except storage.InstanceNotFoundError:
        return "Instance not found", 404


@api.route("/resources/<name>/status", methods=["GET"])
@auth.required
def status(name):
    states = {"started": 204, "pending": 202, "scaling": 204}
    manager = get_manager()
    try:
        status = manager.status(name)
    except storage.InstanceNotFoundError:
        return "Instance not found", 404
    return status, states.get(status, 500)


@api.route("/resources/<name>/scale", methods=["POST"])
@auth.required
def scale_instance(name):
    quantity = request.form.get("quantity")
    if not quantity:
        return "missing quantity", 400
    manager = get_manager()
    try:
        manager.scale_instance(name, int(quantity))
    except ValueError as e:
        msg = " ".join(e.args)
        if "invalid literal" in msg:
            return "invalid quantity: %s" % quantity, 400
        return msg, 400
    except storage.InstanceNotFoundError:
        return "Instance not found", 404
    return "", 201


@api.route("/plugin", methods=["GET"])
def get_plugin():
    return inspect.getsource(plugin)


def register_manager(name, obj, override=False):
    if not override and name in managers:
        raise ValueError("Manager already registered")
    managers[name] = obj


def get_manager():
    manager = os.environ.get("API_MANAGER", "ec2")
    manager_class = managers.get(manager)
    if not manager_class:
        raise ValueError("{0} is not a valid manager".format(manager))
    mongodb_uri = os.environ.get("API_MONGODB_URI")
    mongodb_database = os.environ.get("API_MONGODB_DATABASE_NAME")
    return manager_class(storage.MongoDBStorage(mongo_uri=mongodb_uri,
                                                dbname=mongodb_database))



# Copyright 2014 varnishapi authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import codecs
import httplib2
import os

import varnish
from feaas import storage

VCL_TEMPLATE_FILE = os.path.abspath(os.path.join(os.path.dirname(__file__), "..",
                                                 "misc", "default.vcl"))

DUMP_VCL_FILE = os.path.abspath(os.path.join(os.path.dirname(__file__), "..",
                                             "misc", "dump_vcls.bash"))


class BaseManager(object):

    def __init__(self, storage):
        self.storage = storage

    def new_instance(self, name):
        self._check_duplicate(name)
        instance = storage.Instance(name)
        self.storage.store_instance(instance)
        return instance

    def _check_duplicate(self, name):
        try:
            self.storage.retrieve_instance(name=name)
            raise storage.InstanceAlreadyExistsError()
        except storage.InstanceNotFoundError:
            pass

    def bind(self, name, app_host):
        instance = self.storage.retrieve_instance(name=name)
        bind = storage.Bind(app_host, instance)
        self.storage.store_bind(bind)

    def unbind(self, name, app_host):
        instance = self.storage.retrieve_instance(name=name)
        for unit in instance.units:
            self.remove_vcl(unit.dns_name, unit.secret)
        bind = storage.Bind(app_host, instance)
        self.storage.remove_bind(bind)

    def write_vcl(self, instance_addr, secret, app_addr):
        vcl = self.vcl_template() % {"app_host": app_addr}
        try:
            handler = varnish.VarnishHandler("{0}:6082".format(instance_addr),
                                             secret=secret)
            handler.vcl_inline("feaas", vcl.encode("iso-8859-1", "ignore"))
            handler.vcl_use("feaas")
            handler.quit()
        except AssertionError as e:
            if len(e.args) > 0 and "106 Already a VCL program named" in e.args[0]:
                return
            raise e

    def remove_vcl(self, instance_addr, secret):
        handler = varnish.VarnishHandler("{0}:6082".format(instance_addr),
                                         secret=secret)
        handler.vcl_use("boot")
        handler.vcl_discard("feaas")
        handler.quit()

    def vcl_template(self):
        with codecs.open(VCL_TEMPLATE_FILE, encoding="utf-8") as f:
            content = f.read()
            content = content.replace("\n", " ")
            content = content.replace('"', r'\"')
            content = content.replace("\t", "")
            return '"%s"' % content.strip()

    def remove_instance(self, name):
        instance = self.storage.retrieve_instance(name=name)
        instance.state = "removed"
        self.storage.store_instance(instance)

    def info(self, name):
        instance = self.storage.retrieve_instance(name=name)
        return [{"label": "Address",
                 "value": instance.units[0].dns_name}]

    def status(self, name):
        instance = self.storage.retrieve_instance(name=name)
        return instance.state

    def scale_instance(self, name, quantity):
        if quantity < 1:
            raise ValueError("quantity must be a positive integer")
        instance = self.storage.retrieve_instance(name=name)
        if instance.state == "scaling":
            raise ValueError("instance is already scaling")
        if quantity == len(instance.units):
            raise ValueError("instance already have %d units" % quantity)
        self.storage.store_scale_job({"instance": name, "quantity": quantity,
                                      "state": "pending"})

    def get_user_data(self, secret):
        if "USER_DATA_URL" in os.environ:
            url = os.environ.get("USER_DATA_URL")
            h = httplib2.Http()
            _, user_data = h.request(url)
            return user_data.replace("VARNISH_SECRET_KEY", secret)
        user_data_lines = None
        packages = os.environ.get("API_PACKAGES")
        if packages:
            user_data_lines = ["apt-get update",
                               "apt-get install -y {0}".format(packages),
                               "sed -i -e 's/-T localhost:6082/-T :6082/' /etc/default/varnish",
                               "sed -i -e 's/-a :6081/-a :8080/' /etc/default/varnish",
                               "echo {0} > /etc/varnish/secret".format(secret),
                               "service varnish restart",
                               "cat > /etc/cron.hourly/dump_vcls <<'END'",
                               open(DUMP_VCL_FILE).read(),
                               "END",
                               "chmod +x /etc/cron.hourly/dump_vcls"]
        if user_data_lines:
            return "\n".join(user_data_lines) + "\n"

    def start_instance(self, name):
        raise NotImplementedError()

    def terminate_instance(self, name):
        raise NotImplementedError()

    def physical_scale(self, instance, quantity):
        raise NotImplementedError()
